import fractalartmaker as fam
import turtle


def main():
    fam.demo_four_corners()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_spiral_squares()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_double_spiral_squares()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_triangle_spiral()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_glider()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_sierpinski_triangle()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_wave()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_horn()
    input('Press enter for next demo...')
    turtle.reset()
    
    fam.demo_snowflake()
    input('Press enter for next demo...')
    turtle.reset()
    

if __name__ == '__main__':
    main()
